/*
  */

#include "ocr.h"

#include <X11/Xlib.h>
#include <X11/extensions/XTest.h>
#include <X11/keysym.h>
#include <X11/Xutil.h>

#include <X11/xpm.h>

CanonImage *page_images[MAX_PAGES];
int npages;

static Window id;
static Display *disp;
static struct timeval tv_startup;
static unsigned wwidth, wheight;

static XpmAttributes xpma;

static KeyCode keycode(KeySym sym) {
  return XKeysymToKeycode(disp,sym);
}

#if 0
static void check_exitstatus(int st) {
  eassert(WIFEXITED(st));
  eassert(!WEXITSTATUS(st));
}

static void check_pclose(FILE *f, char *cmd) {
  int r;
  eassert(!ferror(f));
  r= fgetc(f);  eassert(r==EOF);  eassert(feof(f));
  r= pclose(f);  eassert(r>=0);  check_exitstatus(r);
  free(cmd);
}

static CanonImage *screenshot_now(void) {
  char *cmd;
  CanonImage *ci;
  int r;
  
  r= asprintf(&cmd, "xwd -silent -id 0x%lx | xwdtopnm", (unsigned long)id);
  eassert(r>=0);
  FILE *f= popen(cmd,"r");  eassert(f);
  ci= file_read_image(f);
  check_pclose(f, cmd);
  return ci;
}
#endif

static void screenshot_startup(void) {
  int r;
  disp= XOpenDisplay(0);  eassert(disp);
  r= gettimeofday(&tv_startup,0);  eassert(!r);
}

#if 0
static CanonImage *single_page(void) {
  int r;
  r= XRaiseWindow(disp, id);  eassert(r);
  r= XSync(disp, False);  eassert(r);
  return screenshot_now();
}
#endif

/*---------- pager ----------*/

typedef XImage Snapshot;

//static size_t snapshot_alloc= 1024;
static double last_input;
static const double min_update_allowance= 0.5;

static double timestamp(void) {
  struct timeval tv;
  int r;
  
  r= gettimeofday(&tv,0);  eassert(!r);
  double t= (tv.tv_sec - tv_startup.tv_sec) +
            (tv.tv_usec - tv_startup.tv_usec) * 1e-6;
  fprintf(stderr,"%f\n",t);
  return t;
}
static void delay(double need_sleep) {
  int r;
  fprintf(stderr,"PAGING     delay %f\n",need_sleep);
  r= usleep(need_sleep * 1e6);  eassert(!r);
}

static void sync_after_input(void) {
  int r;
  r= XSync(disp, False);  eassert(r);
  last_input= timestamp();
}

static void send_key(KeySym sym) {
  XTestFakeKeyEvent(disp, keycode(sym),1, 10);
  XTestFakeKeyEvent(disp, keycode(sym),0, 10);
}

static void send_pgup_many(void) {
  int i;
  for (i=0; i<25; i++)
    send_key(XK_Prior);
  fprintf(stderr,"PAGING   PageUp x %d\n",i);
  sync_after_input();
}
static void send_pgdown(void) {
  send_key(XK_Next);
  fprintf(stderr,"PAGING   PageDown\n");
  sync_after_input();
}

static void free_snapshot(Snapshot **io) {
  if (*io) XDestroyImage(*io);
  *io= 0;
}

static void snapshot(Snapshot **output) {
//  char *cmd;
//  int r;
//  XImage *xim;
  
  free_snapshot(output);

  fprintf(stderr,"PAGING   snapshot\n");

  timestamp();
  *output= XGetImage(disp,id, 0,0, wwidth,wheight, AllPlanes, ZPixmap);
  timestamp();
  
  fprintf(stderr,"PAGING   snapshot done.\n");
}

static int identical(const Snapshot *a, const Snapshot *b) {
  if (!(a->width == b->width &&
	a->height == b->height &&
	a->bytes_per_line == b->bytes_per_line &&
	a->format == b->format))
    return 0;
  return !memcmp(a->data, b->data, a->bytes_per_line * a->height);
}

static void wait_for_stability(Snapshot **output,
			       const Snapshot *previously,
			       void (*with_keypress)(void)) {
  Snapshot *last=0;
  /* waits longer if we're going to return an image identical to previously
   * if previously==0, all images are considered identical to it */

  fprintf(stderr,"PAGING  wait_for_stability"
	  "  last_input=%f previously=%p\n",
	  last_input, previously);

  for (;;) {
    double at_snapshot= timestamp();
    double need_sleep= min_update_allowance - (at_snapshot - last_input);
    if (need_sleep > 0) { delay(need_sleep); continue; }

    snapshot(output);

    if (!with_keypress &&
	!(previously && identical(*output,previously))) {
      fprintf(stderr,"PAGING  wait_for_stability  simple\n");
      break;
    }

    if (last && identical(*output,last)) {
      fprintf(stderr,"PAGING  wait_for_stability  stabilised\n");
      break;
    }
    
    fprintf(stderr,"PAGING  wait_for_stability  retry\n");

    free_snapshot(&last); last=*output; *output=0;

    if (with_keypress)
      with_keypress();

    delay(0.5);
  }

  free_snapshot(&last);
  fprintf(stderr,"PAGING  wait_for_stability done.\n");
}

static void raise_and_set_focus(void) {
  int r;
  XWindowAttributes attr;
  int xpos,ypos, evbase,errbase,majver,minver;
  unsigned bd,depth;
  Window dummy;
  
  fprintf(stderr,"PAGING raise_and_set_focus\n");

  r= XTestQueryExtension(disp, &evbase,&errbase,&majver,&minver);
  eassert(r==True);

  r= XRaiseWindow(disp, id);  eassert(r);

  r= XGetWindowAttributes(disp, id, &attr);  eassert(r);
  r= XGetGeometry(disp,id, &attr.root,
		  &xpos,&ypos, &wwidth,&wheight,
		  &bd,&depth);
  eassert(r);

  r= XTranslateCoordinates(disp, id,attr.root, 160,160, &xpos,&ypos,
			   &dummy);
  eassert(r);

  int screen= XScreenNumberOfScreen(attr.screen);
  XTestFakeMotionEvent(disp,screen, xpos, ypos, 0);

  XTestFakeButtonEvent(disp,1,1, 50);
  XTestFakeButtonEvent(disp,1,0, 50);

  sync_after_input();
  fprintf(stderr,"PAGING raise_and_set_focus done.\n");
}

static void store_page(int pageno, Snapshot *sn) {
  pid_t converter, paster;
  eassert(pageno < MAX_PAGES);
  int paste[2], results[2];
  FILE *err;
  int r;

  eassert(!fflush(stdout));
  eassert(!fflush(stderr));
  
  r= pipe(paste);  eassert(!r);
  r= pipe(results);  eassert(!r);
  err= tmpfile();  eassert(err);

  converter= fork();  eassert(converter!=-1);
  if (!converter) {
    r= dup2(paste[0],0);  eassert(r==0);
    r= dup2(results[1],1);  eassert(r==1);
    r= dup2(2,4);            eassert(r==4); /* fileno(errn) > 4, see above */
    r= dup2(fileno(err),2);  eassert(r==2);
    close(paste[0]);
    close(paste[1]);
    close(results[0]);
    close(results[1]);
    execlp("xpmtoppm", "xpmtoppm",(char*)0);
    dup2(4,2);
    eassert(!"xpmtoppm exec failure");
  }

  char *xpmdata=0;
  r= XpmCreateBufferFromImage(disp, &xpmdata, sn, 0, &xpma);
  eassert(!r);
  eassert(xpmdata);

  paster= fork();  eassert(paster!=-1);
  if (!paster) {
    FILE *f= fdopen(paste[1],"w");  eassert(f);
    close(paste[0]);
    close(results[0]);
    close(results[1]);
    r= fputs(xpmdata,f);  eassert(r!=EOF);
    //size_t did= fwrite(sn->d, 1, sn->len, f);
    //eassert(did==sn->len);
    eassert(!fclose(f));
    exit(0);
  }

  close(paste[0]);
  close(paste[1]);
  close(results[1]);

  XpmFree(xpmdata);
  
  FILE *f= fdopen(results[0],"r");
  int c1= fgetc(f);
  if (c1!=EOF) {
    ungetc(c1,f);
    page_images[pageno]= file_read_image(f);
    r= fgetc(f);  eassert(r==EOF);  eassert(!ferror(f));  eassert(feof(f));
    fclose(f);
  }

  pid_t got_conv,got_paste;
  int st_conv, st_paste;

  got_conv= waitpid(converter,&st_conv,0);  eassert(got_conv==converter);
  got_paste= waitpid(paster,&st_paste,0);  eassert(got_paste==paster);

  if (!st_conv &&
      (!st_paste || (WIFSIGNALED(st_paste) && WTERMSIG(st_paste)==SIGPIPE))
      && c1!=EOF) {
    fclose(err);
    return;
  }
  rewind(err); int c; while ((c=getc(err))!=EOF) fputc(c,stderr);
  fprintf(stderr, "convert: subprocess statuses: %d %d\n", st_conv, st_paste);
  _exit(127);
}

static void read_pages(void) {
  Snapshot *current=0, *last=0;

  raise_and_set_focus();

  /* page to the top - keep pressing page up until the image stops changing */
  wait_for_stability(&current,0, send_pgup_many);

  /* now to actually page down */
  for (;;) {
    fprintf(stderr,"PAGING page %d\n",npages);
    store_page(npages, current);
    free_snapshot(&last); last=current; current=0;
    fprintf(stderr,"PAGING page %d converted\n",npages);

    wait_for_stability(&current,last, 0);
    if (npages &&  /* first pagedown doesn't do much */
	identical(current,last)) {
      free_snapshot(&current);
      break;
    }

    send_pgdown();
    npages++;
  }
  fprintf(stderr,"PAGING all done.\n");
}    

int main(int argc, char **argv) {
  screenshot_startup();

  id= strtoul(*++argv,0,0);

  read_pages();
  return 0;
}

